"""
Code for identifying variants in illumina reads, based on Gerton's haplotype realignment
algorithm and initial implementation.
"""

from __future__ import division

import logging
import bamfileutils
import pyvcf
import multiprocessing
import variantcaller
import extendedoptparse
import os
import random
import heapq

from variantcaller import PlatypusSingleProcess
from variantcaller import PlatypusMultiProcess

###################################################################################################

logger = logging.getLogger("Log")

###################################################################################################

class FileForQueueing(object):
    """
    """
    def __init__(self, theFile, line):
        """
        Store the file, and initialise the current value
        """
        self.theFile = theFile
        self.line = line
        cols = self.line.strip().split("\t")
        self.chrom = cols[0]

        # Where possible, convert chromosome names into
        # integers for sorting. If not possible, use
        # original names.
        try:
            self.chrom = int(self.chrom.upper().strip("CHR"))
        except:
            pass

        self.pos = int(cols[1])

    def __cmp__(self, other):
        """
        Comparison function. Utilises the comparison function defined in
        the AlignedRead class.
        """
        return cmp(self.chrom, other.chrom) or cmp(self.pos, other.pos)

    def __del__(self):
        """
        Destructor
        """
        self.theFile.close()
        os.remove(self.theFile.name)

    def next(self):
        """
        Increment the iterator and yield the new value. Also, store the
        current value for use in the comparison function.
        """
        self.line = self.theFile.next()
        cols = self.line.strip().split("\t")
        self.chrom = cols[0]

        # Where possible, convert chromosome names into
        # integers for sorting. If not possible, use
        # original names.
        try:
            self.chrom = int(self.chrom.upper().strip("CHR"))
        except:
            pass

        self.pos = int(cols[1])

###################################################################################################

def regionSort(x, y):
    """
    Sort chromosomal regions
    """
    chrom1 = x[0]
    chrom2 = y[0]
    pos1 = int(x[1])
    pos2 = int(y[1])

    try:
        chrom1 = int(chrom1.replace("chr", ""))
        chrom2 = int(chrom2.replace("chr", ""))
    except ValueError:
        pass

    return cmp(chrom1, chrom2) or cmp(pos1, pos2)

###################################################################################################

def chromAndPosSort(x, y):
    """
    Comparison function for use in sort routines. Compares strings of the form
    chr10:0-100. Sorting is done first by chromosome, in alphabetical order, and then
    by start position in numerical order.
    """
    xChrom = x.split("_")[-1].split(":")[0]
    yChrom = y.split("_")[-1].split(":")[0]
    xStart = int(x.split(":")[1].split("-")[0])
    yStart = int(y.split(":")[1].split("-")[0])

    try:
        xChrom = int(xChrom.replace("chr", ""))
        yChrom = int(yChrom.replace("chr", ""))
    except ValueError:
        pass

    return cmp(xChrom, yChrom) or cmp(xStart, yStart)

###################################################################################################

def printRegions(args):
    """
    Split the reference sequence into sensibly-sized regions, and
    print them out.
    """
    parser = extendedoptparse.OptionParser()

    parser.add_option("--refFile",dest="refFile", help="Fasta file of reference. Index must be in same directory", action='store', type='string', required=True)
    parser.add_option("--regions", dest="regions", type="list", help = "region as comma-separated list of chr:start-end, or just list of chr, or nothing", default=None, required=False, action = 'store')
    parser.add_option("--bamFiles", dest="bamFiles", type="list", help = "Comma-delimited list of bam file names", default=None, required=True)
    parser.add_option("--bufferSize", dest="bufferSize", type="int", help = "Max size of region to process. Chromosomes will be broken up into regions of this size", default=100000, required=False)
    parser.add_option("--parseNCBI", dest="parseNCBI", help="", type=int, action='store', default=0)

    (options, args) = parser.parse_args(args)
    regions = bamfileutils.getRegions(options)

    for region in regions:
        print "%s:%s-%s" %(region[0],region[1],region[2])

###################################################################################################

def runVariantCaller(options):
    """
    Run the indel caller

    We feed in a series of windows, each of which has a list of all the possible haplotypes in
    that window. For N P-ploid individuals, there are NP haplotypes to choose and we calculate
    the collection of NP haplotypes which maximise the likelihood.
    """
    # Seed the Python random number generator
    random.seed("FullManyAFlowerIsBornToBlushUnseenAndWasteItsSweetnessOnTheDesertAir")

    # Set up basic logging
    import logging
    import logging.handlers

    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    log = logging.getLogger('Log')

    ch = logging.StreamHandler()
    fh = logging.FileHandler(options.logFileName, 'w')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)

    if options.verbosity == 0:
        log.setLevel(logging.ERROR)
        ch.setLevel(logging.ERROR)
        fh.setLevel(logging.ERROR)
    elif options.verbosity == 1:
        log.setLevel(logging.WARNING)
        ch.setLevel(logging.WARNING)
        fh.setLevel(logging.WARNING)
    elif options.verbosity == 2:
        log.setLevel(logging.DEBUG)
        ch.setLevel(logging.INFO)
        fh.setLevel(logging.DEBUG)
    elif options.verbosity >= 3:
        # Debug goes to file only.
        log.setLevel(logging.DEBUG)
        ch.setLevel(logging.INFO)
        fh.setLevel(logging.DEBUG)
    else:
        raise StandardError, "Value of 'verbosity' input parameter must be between 0 and 3 inclusive"

    log.addHandler(ch)
    log.addHandler(fh)
    logger.info("Beginning variant calling")
    logger.info("Output will go to %s" %(options.output))

    regions = sorted(bamfileutils.getRegions(options), cmp=regionSort)

    if options.nCPU == 1:
        p1 = PlatypusSingleProcess(options.output, options, regions)
        p1.run()

    else:
        # Create process manager
        fileNames = set()
        processes = []
        regionsForEachProcess = []

        for i in range(options.nCPU):
            regionsForEachProcess.append([])

        for index,region in enumerate(regions):
            regionsForEachProcess[index % options.nCPU].append(region)

        for index in range(options.nCPU):
            fileName = options.output + "_temp_%s" %(index)
            fileNames.add(fileName)
            processes.append(PlatypusMultiProcess(fileName, options, regionsForEachProcess[index]))

        for process in processes:
            process.start()

        for process in processes:
            process.join()

        # Final output file
        outputVCF = open(options.output, 'wb')
        theHeap = []

        # Initialise queue
        for index, fileName in enumerate(fileNames):
            theFile = open(fileName, 'rb')

            for line in theFile:

                # End of this file
                if line[0] == "#":
                    if index == 0:
                        outputVCF.write(line)
                    else:
                        continue
                else:
                    theFileForQueueing = FileForQueueing(theFile, line)
                    heapq.heappush(theHeap, theFileForQueueing)
                    break

        # Merge-sort the output using a priority queue
        while len(theHeap) != 0:

            # Get file from heap in right order
            nextFile = heapq.heappop(theHeap)
            outputVCF.write(nextFile.line)

            # Put file back on heap
            try:
                nextFile.next()
                heapq.heappush(theHeap, nextFile)
            except StopIteration:
                continue

        # Close final output file
        outputVCF.close()

    # All done. Write a message to the log, so that it's clear when the
    # program has actually finished, and not crashed.
    logger.info("Finished variant calling")

###################################################################################################

def callVariants(args):
    """
    Run the Platypus variant-caller, with the specified arguments
    """
    parser = extendedoptparse.OptionParser()

    parser.add_option("-o", "--output", dest="output",  help="Output SNP data file", action='store', type='string', default="AllVariants.vcf")
    parser.add_option("-n", "--nIndividuals", dest="nInd", help="Number of Individuals sequenced", action='store', type='int', default=1)
    parser.add_option("-p", "--ploidy", dest="ploidy", help="Number of copief of each chromosome per individual", action='store', type='int', default=2)
    parser.add_option("--refFile",dest="refFile", help="Fasta file of reference. Index must be in same directory", action='store', type='string', required=True)
    parser.add_option("--regions", dest="regions", type="list", help = "region as comma-separated list of chr:start-end, or just list of chr, or nothing", default=None, action = 'store')
    parser.add_option("--bamFiles", dest="bamFiles", type="list", help = "Comma-delimited list of bam file names", default=None, required=True)
    parser.add_option("--bufferSize", dest="bufferSize", type="int", help = "Data will be buffered in regions of this size", default=1000000, required=False)
    parser.add_option("--minReads", dest="minReads", help="Minimum required number of reads in window", action='store', type='int', default=2)
    parser.add_option("--maxHaplotypes", dest="maxHaplotypes", help="Maximium haplotypes to consider in a given window", action='store', type='int', default=256)
    parser.add_option("--maxVariants", dest="maxVariants", help="Maximium variants to consider in a given window", action='store', type='int', default=8)
    parser.add_option("--maxReads", dest="maxReads", help="Maximium coverage in window", action='store', type='float', default=5000000)
    parser.add_option("--maxEMIterations", dest="maxEMIterations", help="Maximium number of iterations for EM algorithm", action='store', type='int', default=100)
    parser.add_option("--maxSize", dest="maxSize", help="Largest indel to consider", action='store', type='int', default=250)
    parser.add_option("--verbosity", dest="verbosity", help="Level of logging", action='store', type='int', default=2)
    parser.add_option("--minMapQual", dest="minMapQual", help="Minimum mapping quality of read. Any reads with map qual below this are ignored", action='store', type = 'int', default=20, required=False)
    #parser.add_option("--maxWindowSize", dest="maxWindowSize", help="Maximum size of genome window used for variant calling", action='store', type = 'int', default=250, required=False)
    parser.add_option("--maxBadQualBases", dest="maxBadQualBases", help="Max number of bases per read that can have base-quality <= 20", action='store', type = 'int', default=20, required=False)
    parser.add_option("--minGoodQualBases", dest="minGoodQualBases", help="Minimum number of bases per read that must have base-quality >= 20", action='store', type = 'int', default=20, required=False)
    parser.add_option("--minBaseQual", dest="minBaseQual", help="Minimum allowed base-calling quality. Any bases with qual below this are ignored in SNP-calling", action='store', type = 'int', default=20, required=False)
    parser.add_option("--maxReadLength", dest="rlen", help="Maximum read length", action='store', type = 'int', default=100)
    parser.add_option("--minFlank", dest="minFlank", help="Flank size for indel candidates", action='store', type = 'int', default=3)
    parser.add_option("--logFileName", dest="logFileName", help="Name of log file", action='store', type='string', default="log.txt")
    parser.add_option("--source", dest="sourceFile", help="vcf file(s) to get candidates from", action='store', type='list', default= None)
    parser.add_option("--nCPU", dest="nCPU", help="Number of processors to use", action='store', type='int', default=1)
    parser.add_option("--numPasses", dest="numPasses", help="Number of passes to do through variant windows", action='store', type='int', default=1)
    parser.add_option("--getVariantsFromBAMs", dest="getVariantsFromBAMs", help="If set to TRUE (default), variant candidates will be generated from BAMs as well as any other inputs", action='store', type='int', default=1)
    parser.add_option("--genSNPs", dest="genSNPs", help="If set to TRUE (default), SNP candidates will be considered", action='store', type='int', default=1)
    parser.add_option("--genIndels", dest="genIndels", help="If set to TRUE (default), Indel candidates will be considered", action='store', type='int', default=1)
    parser.add_option("--mergeClusteredVariants", dest="mergeClusteredVariants", help="If set to 1, variant-containing windows which are close together will be merged, resulting in slower, more accurate variant calls in diverse regions", action='store', type='int', default=0)
    parser.add_option("--maxVarDist", dest="maxVarDist", help="Max distance between variants to be considered in the same window", action='store', type='int', default=25)
    parser.add_option("--parseNCBI", dest="parseNCBI", help="", type=int, action='store', default=0)
    parser.add_option("--useIndelErrorModel", dest="useIndelErrorModel", help="If set to 1, the new indel error model will be used", action='store', type='int', default=0)
    parser.add_option("--minPosterior", dest="minPosterior", help="Only variants with posterior >= this will be outpu to the VCF. Value is a Phred-score.", action='store', type='int', default=5)
    parser.add_option("--sbThreshold", dest="sbThreshold", help="P-value for strand-bias filtering..", action='store', type='float', default=1e-2)
    parser.add_option("--abThreshold", dest="abThreshold", help="P-value for allele-bias filtering..", action='store', type='float', default=1e-5)
    parser.add_option("--minVarFreq", dest="minVarFreq", help="Variants below this frequency will be flagged as allele-biased", action='store', type='float', default=0.2)
    parser.add_option("--useCortex", dest="useCortex", help="If 1, Cortex will be used to assemble variant candidates for Platypus to call.", action='store', type='int', default=0)
    parser.add_option("--useEMLikelihoods", dest="useEMLikelihoods", help="If 1, likelihoods computed from EM algorithm will be used to call genotypes for each sample, otherwise likelihoods from individual sample will be used.", action='store', type='int', default=0)
    #parser.add_option("--useHapFilter", dest="useHapFilter", help="If 1, Platypus filters variants in difficult regions using alignment scores of reads to multi-variant haplotypes.", action='store', type='int', default=0)
    parser.add_option("--filterVarsByCoverage", dest="filterVarsByCoverage", help="If 1, Platypus filters variants in difficult regions by the number of times each variant is seen.", action='store', type='int', default=1)
    parser.add_option("--badReadsWindow", dest="badReadsWindow", help="Size of window around variant to look for low-quality bases.", action='store', type='int', default=11)
    parser.add_option("--badReadsThreshold", dest="badReadsThreshold", help="Variants where the median minimum quality in a window of badReadsWindow around the variant position falls below this value will be filtered with the flag 'badReads'.", action='store', type='int', default=15)
    parser.add_option("--filteredReadsFrac", dest="filteredReadsFrac", help="If > this fraction of reads are filtered in a given window, the 'badReads filter is triggered.", action='store', type='float', default=0.7)

    (options, args) = parser.parse_args(args)
    runVariantCaller(options)

###################################################################################################
