"""
This module contains utilities for generating a list windows, and the reads
and variant candidates they contain, for processing by the Platypus variant
caller.
"""
from __future__ import division

import logging
import variant
import variantutils

from variant import Variant,VariantCandidateGenerator

logger = logging.getLogger("Log")

###################################################################################################

class WindowGenerator(object):
    """
    This class is used to generate a list of all possible variants for a given region.
    """
    def __init__(self):
        """
        Constructor. Currently does nothing.
        """
        pass

    def getVariantsByPos(self, chromosome, start, end, sortedVariants):
        """
        Returns a list of lists of variant candidates. Each list contains all candidates
        at that position.
        """
        varsByPos = {}

        for v in sortedVariants:
            if v.refName == chromosome and v.refPos >= start and v.refPos <= end:
                try:
                    varsByPos[v.refPos].append(v)
                except KeyError:
                    varsByPos[v.refPos] = [v]

        listOfVarsByPos = []

        for pos in sorted(varsByPos.keys()):
            listOfVarsByPos.append(varsByPos[pos])

        return listOfVarsByPos

    def getBunchesOfInteractingVariants(self, varsByPos, options):
        """
        Go through the list of lists or variants sorted by position, and
        concatenate neighbouring lists, if any of the variants in those
        lists interact.
        """
        bunchedVars = []

        for varList in varsByPos:

            if len(bunchedVars) == 0:
                bunchedVars.append(varList)
            else:
                minLastPos = bunchedVars[-1][0].minRefPos
                minThisPos = varList[0].minRefPos
                maxLastPos = bunchedVars[-1][-1].maxRefPos
                maxThisPos = varList[-1].maxRefPos
                maxIndelPosLastWindow = -1
                minIndelPosThisWindow = -1

                for var in reversed(bunchedVars[-1]):
                    if var.nAdded != var.nRemoved:
                        maxIndelPosLastWindow = var.maxRefPos
                        break # Going backwards, max will always be the first

                for var in varList:
                    if var.nAdded != var.nRemoved:
                        minIndelPosThisWindow = var.minRefPos
                        break # Min will always be the first one

                # Always put interacting variants, or variants closer than
                # one read length, in the same window.
                if maxLastPos >= minThisPos:
                    bunchedVars[-1].extend(varList)

                # Try to merge windows, if edge variants are close together, and resulting window
                # size is not too large, and the mergeClusteredVariants option is set
                elif options.mergeClusteredVariants:

                    totalVars = len(varList) + len(bunchedVars[-1])
                    maxVars = options.maxVariants
                    readLength = options.rlen
                    maxWindowSize = readLength
                    thisWindowSize = maxThisPos - minLastPos
                    edgeVarDist = minThisPos - maxLastPos

                    if maxIndelPosLastWindow == -1 and minIndelPosThisWindow == -1:
                        closestIndelDist = 100*readLength
                    elif maxIndelPosLastWindow == -1:
                        closestIndelDist = minIndelPosThisWindow - maxLastPos
                    elif minIndelPosThisWindow == -1:
                        closestIndelDist = minThisPos - maxIndelPosLastWindow
                    else:
                        closestIndelDist = min(minIndelPosThisWindow - maxLastPos, minThisPos - maxIndelPosLastWindow)

                    #if thisWindowSize <= maxWindowSize and edgeVarDist < readLength and totalVars <= maxVars:
                    #if thisWindowSize <= maxWindowSize and (closestIndelDist < readLength or edgeVarDist < readLength/2):
                    if thisWindowSize <= maxWindowSize and edgeVarDist < options.maxVarDist:
                        bunchedVars[-1].extend(varList)
                    else:
                        bunchedVars.append(varList)

                # If there is no interaction, start a new window
                else:
                    bunchedVars.append(varList)

        return bunchedVars

    def getWindowVariants(self, chromosome, start, end, sortedVariants, options):
        """
        Iterate over the list of bunches of interacting vars, and concatenate
        neighbouring bunches, if the total number of variants in the two neighbouring
        bunches is <= maxVarsInWindow. Return a list of lists, where each sub-list
        is a set of variants to be considered in the same window.
        """
        varsByPos = self.getVariantsByPos(chromosome, start, end, sortedVariants)
        bunchedVars = self.getBunchesOfInteractingVariants(varsByPos, options)
        return bunchedVars

    def WindowsAndVariants(self, chromosome, start, end, sortedVariants, options):
        """
        Generator: return a series of dictionaries, each containing the start and end
        positions of a haplotype window, plus a list of all the variants in that window.
        Basically takes a stream of variant candidates, and figures out where to break
        them up into windows and then gives that to the haplotype generator.

        Yields a dictionary.

        The basic procedure is as follows: loop through all variants. If the next variant is outside the
        window, then either extend the window, or, if we already have enough variants in the current
        window, then yield this window, and start a new one which will contain the next variant.

        The maximum allowed number of variants in a particular window is currently set to 3.
        Once we hit the end of the variants, we catch the StopIteration, and return whatever is in the the variant buffer.
        """
        windowVars = self.getWindowVariants(chromosome, start, end, sortedVariants, options)

        nVar = len(sortedVariants)
        nSnp = len([v for v in sortedVariants if abs(v.nAdded - v.nRemoved) == 0])

        if options.verbosity >= 3:
            logger.debug("There are %s vars in total in this region" %(nVar))
            logger.debug("There are %s SNPs and %s indels" %(nSnp, nVar-nSnp))
            logger.debug("There will be %s windows used in this region" %(len(windowVars)))

        for index,vars in enumerate(windowVars):

            winStart = max(vars[0].minRefPos,start)
            winEnd = min(vars[-1].maxRefPos,end)

            if winEnd == winStart:
                winEnd = winEnd + 1

            thisWindow = dict(chromosome=chromosome,startPos=winStart, endPos=winEnd, variants=vars)

            if options.verbosity >= 3:
                logger.debug("Window = %s" %(thisWindow))

            yield thisWindow

###################################################################################################
